#include <windows.h>
#include <commctrl.h>
#include <advpub.h>
#include <stdio.h>
#include "libs\GetArgv.h"
#include "libs\SimpleString.h"
#include "resource.h"
#include "version.h"

#ifndef SETUP_LITE
#define MessageBoxWarn(a)  MessageBox(NULL, a, TEXT(SETUP_TITLE_STR), MB_OK | MB_ICONWARNING)
#define MessageBoxInfo(a)  MessageBox(NULL, a, TEXT(SETUP_TITLE_STR), MB_OK | MB_ICONINFORMATION)
#define MessageBoxError(a) MessageBox(NULL, a, TEXT(SETUP_TITLE_STR), MB_OK | MB_ICONERROR)
#else
#define MessageBoxWarn(a)
#define MessageBoxInfo(a)
#define MessageBoxError(a)
#endif

// External dependencies
PVOID __fastcall GetResource( PCTSTR pszName, PCTSTR pszType, PDWORD pcbData );
UINT __fastcall ExtractCabResource( PSTR pszOutputPath, PSTR pszOutputPathAppend );
BOOL DoSetup( PSTR pszPath, PSTR pszPathAppend );

// Internal helpers
__forceinline DWORD SetupMain( HINSTANCE hInstance );
__forceinline BOOL SetupRun( HINSTANCE hInstance, BOOL bInstall, BOOL bNotify );
__forceinline SIZE_T GetTempNameA( PSTR pszTempName, SIZE_T cbTempName );

#ifndef SETUP_LITE
__forceinline BOOL ShowLicense( HINSTANCE hInstance );
INT_PTR CALLBACK LicenseDlgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
#endif

#pragma comment(linker, "/entry:SetupEntry")
void SetupEntry( )
{
	ExitProcess(SetupMain(GetModuleHandleA(NULL)));
}

/**
 * SetupMain - Process command line arguments, display help, display license
 **/
__forceinline DWORD SetupMain( HINSTANCE hInstance )
{
	int argc = 0;
	PTSTR *argv = GetArgv(&argc);

	BOOL bInstall = TRUE;
	#ifndef SETUP_LITE
	BOOL bNotify = TRUE;
	BOOL bHelp = FALSE;
	#else
	BOOL bNotify = FALSE;
	#endif

	if (argc && argv)
	{
		while (--argc)
		{
			if (lstrcmpi(argv[argc], TEXT("/extract")) == 0)
				bInstall = FALSE;
			#ifndef SETUP_LITE
			else if (lstrcmpi(argv[argc], TEXT("/quiet")) == 0)
				bNotify = FALSE;
			else if (argv[argc][0] == TEXT('/'))
				bHelp = TRUE;
			#endif
		}

		LocalFree(argv);
	}

	#ifndef SETUP_LITE
	InitCommonControls();

	if (bHelp)
	{
		MessageBoxInfo(TEXT("Command-line options:\n/quiet\tSuppress all messages and dialogs\n/extract\tExtract files without installing"));
		return(0);
	}
	else if (bNotify && bInstall && !ShowLicense(hInstance))
	{
		// Exit if the user says "No"
		return(0);
	}
	#endif

	return((SetupRun(hInstance, bInstall, bNotify)) ? 0 : 1);
}

/**
 * SetupRun - Create temp dir, extract cabinet, install, clean-up
 **/
__forceinline BOOL SetupRun( HINSTANCE hInstance, BOOL bInstall, BOOL bNotify )
{
	BOOL bSuccess = FALSE;
	CHAR szPath[MAX_PATH << 1];
	PSTR pszPathAppend = szPath;

	if (bInstall)
	{
		// Installation: Extract to temp directory, call DoSetup, and clean up
		// if necessary

		SIZE_T cbPathRoot = GetTempNameA(szPath, sizeof(szPath));
		pszPathAppend += cbPathRoot;

		SSCpy2ChA(pszPathAppend, '\\', 0);
		++pszPathAppend;

		if (cbPathRoot && ExtractCabResource(szPath, pszPathAppend))
		{
			#ifdef REBOOT_MESSAGE
			DWORD dwRebootCookie = NeedRebootInit();
			#endif

			if (DoSetup(szPath, pszPathAppend))
			{
				if (bNotify)
				{
					MessageBoxInfo(TEXT("Installation complete!"));

					#ifdef REBOOT_MESSAGE
					if (NeedReboot(dwRebootCookie))
						MessageBoxInfo(REBOOT_MESSAGE);
					#endif
				}

				bSuccess = TRUE;
			}
			else if (bNotify)
			{
				MessageBoxWarn(TEXT("Installation error encountered"));
			}
		}
		else if (bNotify)
		{
			MessageBoxError(TEXT("Unexpected file error"));
		}

		if (cbPathRoot)
		{
			SHFILEOPSTRUCTA fileop;
			ZeroMemory(&fileop, sizeof(fileop));
			fileop.wFunc = FO_DELETE;
			fileop.pFrom = szPath;
			fileop.fFlags = FOF_SILENT | FOF_NOCONFIRMATION | FOF_NOERRORUI;

			// We want a double-null termination so that SHFileOperation is happy.
			SSCpy2ChA(--pszPathAppend, 0, 0);

			SHFileOperationA(&fileop);
		}
	}
	else
	{
		// Extraction: Just extract to the current location; no need for temp

		UINT uExtracted;

		SSCpy4ChA(szPath, '.', '\\', 0, 0);
		pszPathAppend += 2;

		if (uExtracted = ExtractCabResource(szPath, pszPathAppend))
		{
			#ifndef SETUP_LITE
			if (bNotify)
			{
				TCHAR szMessage[0x30];
				wsprintf(szMessage, TEXT("%u file(s) successfully extracted"), uExtracted);
				MessageBoxInfo(szMessage);
			}
			#endif

			bSuccess = TRUE;
		}
		else if (bNotify)
		{
			MessageBoxError(TEXT("Unexpected file error"));
		}
	}

	return(bSuccess);
}

/**
 * GetTempNameA - Gets an unused temporary name; does not create file/dir
 **/
__forceinline SIZE_T GetTempNameA( PSTR pszTempName, SIZE_T cbTempName )
{
	PSTR pszBuffer = _tempnam("", "setup");
	SIZE_T cbBuffer;

	if (pszBuffer && (cbBuffer = SSLenA(pszBuffer)) < cbTempName)
	{
		// The null terminator is omitted because we will be appending to
		// this string before using it anyway.
		SSChainNCpyA(pszTempName, pszBuffer, cbBuffer);
		free(pszBuffer);
		return(cbBuffer);
	}

	if (pszBuffer)
		free(pszBuffer);

	return(0);
}

#ifndef SETUP_LITE

/**
 * ShowLicense - Open license dialog
 **/
__forceinline BOOL ShowLicense( HINSTANCE hInstance )
{
	return(IDYES == DialogBoxParam(
		hInstance,
		MAKEINTRESOURCE(IDD_LICENSE),
		NULL,
		LicenseDlgProc,
		(LPARAM)hInstance
	));
}

/**
 * LicenseDlgProc - License dialog UI
 **/
INT_PTR CALLBACK LicenseDlgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch (uMsg)
	{
		case WM_INITDIALOG:
		{
			PSTR pszLicense = GetResource(MAKEINTRESOURCE(IDR_LICENSE), RT_RCDATA, NULL);

			if (pszLicense)
			{
				// Set the window icon
				SendMessage(
					hWnd,
					WM_SETICON,
					ICON_BIG, // No need to explicitly set the small icon
					(LPARAM)LoadIcon((HINSTANCE)lParam, MAKEINTRESOURCE(IDI_SETUP))
				);

				// Set license text
				SetDlgItemTextA(hWnd, IDC_LICENSE_BOX, pszLicense);
			}
			else
			{
				EndDialog(hWnd, IDYES);
			}

			return(TRUE);
		}

		case WM_CLOSE:
		{
			EndDialog(hWnd, IDNO);
			return(TRUE);
		}

		case WM_COMMAND:
		{
			switch (LOWORD(wParam))
			{
				case IDYES:
				case IDNO:
				{
					EndDialog(hWnd, LOWORD(wParam));
					return(TRUE);
				}
			}
		}
	}

	return(FALSE);
}

#endif
